

/**
 * Title: Iterated Functional Systems
 * Description:
 * Copyright:    Copyright (c) 2001
 * Company:
 * @author Michail Popov
 * @version 1.0
 */
import java.io.*;
import java.text.*;
import java.util.*;
import java.awt.*;
import java.awt.event.*;

public class IFS extends Panel implements ComponentListener {
  Button brepaint, badd, bdel, blenses;
  Image image;
  Graphics ig;
  int width, height;
  double[][]trs;
  Vector vtrs;
  Vector names;

  int current;
  int skip, its;
  TextField tskip, tits, tname;
  Choice ctrs;
  TextArea tmat;

  static int border = 20;
  static double def_trs[][][] = new double[][][]{{{0,-0.5,0.5,0,0.5,0},{0,0.5,-0.5,0,0.5,0.5},{0.5,0,0,0.5,0.25,0.5}}};
  static String[] def_names = new String[]{"Crystal"};

  public static void main(String args[]) {
    Frame frame = new Frame("Iterated Functional Systems");
    IFS ifs = new IFS();
    frame.add("Center", ifs);
    frame.pack();
    Dimension d = Toolkit.getDefaultToolkit().getScreenSize();
    frame.setSize(d.width*3/4, d.height*3/4);
    frame.setLocation((d.width - frame.getSize().width) / 2, (d.height - frame.getSize().height) / 2);
    frame.setVisible(true);
  }

  public IFS() {
    Frame f = new Frame("OPTIONS");
    Component c;
    Panel ctl = new Panel();
    GridBagLayout gbl = new GridBagLayout();
    GridBagConstraints gbc = new GridBagConstraints();
    ctl.setLayout(gbl);
    gbc.anchor = gbc.WEST;
    gbc.fill = gbc.HORIZONTAL;
    gbc.weightx = 0.0;
    gbc.weighty = 1.0;
    gbc.gridwidth = 1;

    c = new Label("Skip");
    gbl.setConstraints(c, gbc);
    ctl.add(c);
    gbc.gridwidth = gbc.REMAINDER;
    tskip = new TextField();
    tskip.addTextListener(new TextListener() {
      public void textValueChanged(TextEvent e) {
        try {
	  int skip2 = Integer.parseInt(tskip.getText());
	  if (skip2<=0)
	    throw new Exception();
	  skip = skip2;
        } catch(Exception ex) {
	  tskip.setText(""+skip);
        }
      }
    } );
    c = tskip;
    gbl.setConstraints(c, gbc);
    ctl.add(c);

    gbc.gridwidth = 1;
    c = new Label("Iterations");
    gbl.setConstraints(c, gbc);
    ctl.add(c);
    gbc.gridwidth = gbc.REMAINDER;
    tits = new TextField();
    tits.addTextListener(new TextListener() {
      public void textValueChanged(TextEvent e) {
        try {
	  int its2 = Integer.parseInt(tits.getText());
	  if (its2<=0)
	    throw new Exception();
	  its = its2;
        } catch(Exception ex) {
	  tits.setText(""+its);
        }
      }
    } );

    c = tits;
    gbl.setConstraints(c, gbc);
    ctl.add(c);

    ctrs = new Choice();
    ctrs.addItemListener(new ItemListener() {
      public void itemStateChanged(ItemEvent e) {
        setCurrent(ctrs.getSelectedIndex());
      }
    } );
    c = ctrs;
    gbl.setConstraints(c, gbc);
    ctl.add(c);

    tmat = new TextArea();
    tmat.setFont(new Font("Courier", Font.PLAIN, 12));

    tmat.addTextListener(new TextListener() {
      public void textValueChanged(TextEvent e) {
        boolean ok = getMatrix()!=null;
        badd.setEnabled(ok);
        brepaint.setEnabled(ok);
        blenses.setEnabled(ok);
      }
    } );


    c = tmat;
    gbl.setConstraints(c, gbc);
    ctl.add(c);

    gbc.gridwidth = 1;
    c = new Label("Name");
    gbl.setConstraints(c, gbc);
    ctl.add(c);

    tname = new TextField();
    gbc.weightx = 1;
    c = tname;
    gbl.setConstraints(c, gbc);
    ctl.add(c);
    gbc.weightx = 0;
    gbc.fill = gbc.NONE;
    gbc.anchor = gbc.EAST;
    badd = new Button("Add");
    badd.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent e) {
        String name = tname.getText();
	int index = names.lastIndexOf(name);
	if (index>0) {
	  names.removeElementAt(index);
	  vtrs.removeElementAt(index-1);
	} else
	  index = names.size();
	double[][] d = getMatrix();
	vtrs.insertElementAt(d, index-1);
	names.insertElementAt(name, index);
	updateChoice();
	setCurrent(index);
      }
    } );

    c = badd;
    gbl.setConstraints(c, gbc);
    ctl.add(c);
    bdel = new Button("Del");
    bdel.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent e) {
	int index = ctrs.getSelectedIndex();
	vtrs.removeElementAt(index-1);
	names.removeElementAt(index);
	updateChoice();
	setCurrent(0);
      }
    } );
    c = bdel;
    gbc.gridwidth = gbc.REMAINDER;
    gbl.setConstraints(c, gbc);
    ctl.add(c);

    Panel px = new Panel();
    Button b = new Button("Save All");
    b.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent e) {
        try {
          ObjectOutputStream oos = new ObjectOutputStream(new FileOutputStream("ifs.data"));
          oos.writeObject(vtrs);
	  oos.writeObject(names);
	  oos.writeInt(skip);
	  oos.writeInt(its);
	  oos.close();
        } catch(Exception ex) {
	  System.err.println(ex);
        }
      }
    } );
    px.add(b);
    b = new Button("Repaint");
    brepaint = b;
    b.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent e) {
        update();
      }
    } );
    px.add(b);

    b = new Button("Lenses");
    blenses = b;
    b.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent e) {
        update2();
      }
    } );
    px.add(b);

    b = new Button("Exit");
    b.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent e) {
        System.exit(0);
      }
    } );
    px.add(b);

    gbc.fill = gbc.NONE;
    gbc.anchor = gbc.CENTER;
    gbc.weightx = 1;
    gbc.gridwidth = gbc.REMAINDER;
    c = px;
    gbl.setConstraints(c, gbc);
    ctl.add(c);

    f.add("Center", ctl);
    f.pack();
    f.setVisible(true);


    Panel p1 = new Panel();


    addComponentListener(this);

    try {
      ObjectInputStream ois = new ObjectInputStream(new FileInputStream("ifs.data"));
      vtrs = (Vector)ois.readObject();
      names = (Vector)ois.readObject();
      skip = ois.readInt();
      its = ois.readInt();
      ois.close();
    } catch(Exception e) {
      skip = 500;
      its = 50000;
      vtrs = new Vector();
      names = new Vector();
      names.addElement("none");
      for (int i=0; i<def_trs.length; i++) {
        vtrs.addElement(def_trs[i]);
        names.addElement(def_names[i]);
      }
    }
    tskip.setText(""+skip);
    tits.setText(""+its);

    updateChoice();
    setCurrent(0);
  }

  public void updateChoice() {
    ctrs.removeAll();
    for (int i=0; i<names.size(); i++)
      ctrs.add(names.elementAt(i)+"");
  }

  public double[][] getMatrix() {
    String s = tmat.getText();
    Vector v = new Vector();
    StringTokenizer st = new StringTokenizer(s);
    try {
      while(st.hasMoreTokens()) {
        v.addElement(Double.valueOf(st.nextToken()));
      }
    } catch(Exception e) {
      return null;
    }
    if (v.size()%6!=0||v.size()==0)
      return null;
    double[][]d=new double[v.size()/6][6];
    for (int i=0; i<v.size(); i++)
      d[i/6][i%6]=((Double)v.elementAt(i)).doubleValue();
    return d;
  }


  public void setCurrent(int current) {
    if (ctrs.getSelectedIndex()!=current)
      ctrs.select(current);
    brepaint.setEnabled(current>0);
    blenses.setEnabled(current>0);
    bdel.setEnabled(current>0);
    badd.setEnabled(current>0);
    this.current = current;
    tmat.setText("");
    tname.setText(names.elementAt(current)+"");
    if (current>0) {
      trs = (double[][])vtrs.elementAt(current-1);
      DecimalFormatSymbols dcf = new DecimalFormatSymbols();
      dcf.setDecimalSeparator('.');
      DecimalFormat dc = new DecimalFormat("0.0000", dcf);
      String s="";
      for (int i=0; i<trs.length; i++) {
        for (int j=0; j<trs[i].length; j++) {
	  if (trs[i][j]>=0)
	    s+=" ";
	  s+=dc.format(trs[i][j])+"  ";
        }
	s+="\n";
      }
      tmat.setText(s);
//      update();
    }
  }

  public void componentResized(ComponentEvent e) {
    image = null;
    if (brepaint.isEnabled())
      update();
  }

  public void componentMoved(ComponentEvent e) {

  }

  public void componentHidden(ComponentEvent e) {

  }

  public void componentShown(ComponentEvent e) {

  }

  public void update() {
    System.out.print("update...");
    if (image==null) {
      width = getSize().width;
      height = getSize().height;
      if (width<=0||height<=0)
	return;
      image = createImage(width, height);
      if (image==null)
	return;
      ig = image.getGraphics();
    }
    ig.setColor(Color.white);
    ig.fillRect(0, 0, width, height);
    ig.setColor(Color.black);
    ig.drawRect(border, border, width-2*border, height-2*border);

//-----------------------------
    int w = width-2*border;
    int h = height-2*border;
    trs = getMatrix();
    double x = Math.random();
    double y = Math.random();

    for (int i=0; i<its+skip; i++) {
      double[] t = trs[(int)(Math.random()*trs.length)];
      double ox = x;
      x = x*t[0]+y*t[1]+t[4];
      y = ox*t[2]+y*t[3]+t[5];
      if (i>skip) {
        int px = border+(int)(w*x);
        int py = border+h-(int)(h*y);
        ig.drawLine(px, py, px+1, py+1);
      }
    }
//-----------------------------
    System.out.println("done");
    repaint();
  }

  public void update2() {
    if (image==null) {
      width = getSize().width;
      height = getSize().height;
      if (width<=0||height<=0)
	return;
      image = createImage(width, height);
      if (image==null)
	return;
      ig = image.getGraphics();
    }
    int w = width-2*border;
    int h = height-2*border;
    trs = getMatrix();
    ig.setColor(Color.white);
    ig.fillRect(0, 0, width, height);
    ig.setColor(Color.black);
    ig.drawRect(border, border, width-2*border, height-2*border);

    ig.setColor(Color.red);
    double [] xk = {0,1,1,0, 0, 0.1,0.1,0.5 };
    double [] yk = {0,0,1,1, 0, 0.9,0.1,0.1 };
    double x,y;
    for (int i=0; i<trs.length; i++) {
      double[] t = trs[i];
      int ox = 0;
      int oy = 0;
      for (int j=0; j<xk.length; j++) {
        x = xk[j];
        y = yk[j];
        double ox2 = x;
        x = x*t[0]+y*t[1]+t[4];
        y = ox2*t[2]+y*t[3]+t[5];
        int px = border+(int)(w*x);
        int py = border+h-(int)(h*y);
	if (j>0&&j!=5)
          ig.drawLine(ox, oy, px, py);
	ox = px;
	oy = py;
      }
    }
    repaint();
  }



  public void paint(Graphics g) {
    if (image!=null)
      g.drawImage(image, 0, 0, this);
  }

  public Dimension getPreferredSize() {
    if (image!=null)
      return new Dimension(image.getWidth(this), image.getHeight(this));
    return new Dimension(0, 0);
  }

  public Dimension getMinimumSize() {
    return getPreferredSize();
  }

}